#!/usr/bin/env python
# coding: utf-8

# # Bitterness Classification With XGBoost
# 
# This example has been created as part of our Advanced Topics Webinar, scheduled for May 28, 2021. This webinar focuses on taking existing machine learning workflows and bringing them to the Exabyte platform.
# 
# In this notebook, we will train a model that predicts whether a molecule is bitter or not by using various structural descriptors. The dataset we use is from the BitterSweet dataset hosted in [kaggle](https://www.kaggle.com/katyaarnold/bittersweet). This dataset is actually the combination of several other databases, which are listed in the following publication:
# 
# Tuwani, R., Wadhwa, S. & Bagler, G. BitterSweet: Building machine learning models for predicting the bitter and sweet taste of small molecules. Sci Rep 9, 7155 (2019). https://doi.org/10.1038/s41598-019-43664-y 
#     
# ## Imports
# 
# In addition to a few standard ML packages (pandas, sklearn, etc), we'll also import a few other packages:
#  
# - XGBoost: eXtreme Gradient Boosting. It's a gradient-boosting package that is a lot faster than the sklearn implementation. It also offers the option to apply dropout regularization to the ensemble via the DART algorithm (https://xgboost.readthedocs.io/en/latest/tutorials/dart.html)
# - Imbalanced-Learn: This is a package offering variations on the Synthetic Minority Oversampling Technique (SMOTE) technique, which upsamples the minority class. This is helpful in the case of an imbalanced classification problem.
# - Optuna: This is a hyperparameter optimization library. It implements bayesian optimization to sample the hyperparameter space, in order to maximize a model's predictive capabilities. We use it to tune some of XGBoost's hyperparameters (such as number of trees in the ensemble)
# - RDKit: This is a cheminformatics package focused on molecules; it offers a wide range of descriptors and fingerprints for molecules.
#     

# In[]:


get_ipython().system('pip install numpy matplotlib pandas sklearn xgboost optuna imblearn rdkit_pypi')


# In[]:


import functools

import numpy as np
import matplotlib
import matplotlib.pyplot as plt

import pandas as pd
import sklearn.preprocessing
import sklearn.pipeline

import xgboost 
import optuna
import joblib

import imblearn

from rdkit.Chem import MolFromSmiles, Lipinski
import rdkit.Chem.Descriptors as Descriptors


from rdkit import RDLogger
RDLogger.DisableLog('rdApp.*')


# # Read the Dataset
# Let's begin by reading in the dataset. We have gently cleaned the CSV file, retaining only a few columns:
# 
# - Name: the name of the chemical
# - Taste: Whether the material tastes sweet, bitter, non-bitter, or is tasteless
# - Reference: The dataset the entry is from (see Tuwani et al for further details)
# - SMILES: the SMILES string representing the molecule
# - Canonical SMILES: a Canonical SMILES string (this helps us remove duplicate molecules)
# 
# We can see that the dataset contains comparable numbers of sweet/bitter molecules, along with several that have been labeled as either "tasteless" or "nonbitter"

# In[]:


# Data is from https://www.kaggle.com/katyaarnold/bittersweet
raw_data = pd.read_csv("../assets/bitter-sweet.csv")
print(raw_data.Taste.value_counts())
raw_data


# # Process the Dataset
# We're going to derive all of our properties from the structure of the molecule, which is represented in a SMILES string. We'll take the Canonical SMILES, which is a variant of SMILES that attempts to have only one representation for a molecule. This will help us detect duplicate entries.
# 
# Also, since binary classification is an easier problem, we'll take the all-versus-one approach to labeling. Specifically, we're going to predict whether a material is bitter or non-bitter. To do this, we'll use sklearn's built-in label binarizer, which takes in a list of classes and returns a numpy array containing a one-hot encoding of whether a row is a member of a given class. We'll then take the column for "Bitter" and append it to the pandas array.
# 
# Finally, we'll drop any missing entries from the dataset, and we'll then drop any duplicate rows (judged either by the name or canonical simles).

# In[]:


# Drop a few columns we don't want
data = raw_data[["Name", "Canonical SMILES"]].copy()

# Let's take a 1 v all approach to this, and just encode "Bitter" as 1 and everything else as 0.
classes = ("Sweet", "Bitter", "Tasteless", "Non-bitter")
binarized = sklearn.preprocessing.label_binarize(raw_data.Taste.to_numpy(), classes=classes)
bitter = binarized[:,classes.index("Bitter")]

data["Bitter"] = bitter
data = data.dropna()
for colname in data.drop(columns=["Bitter"]).columns:
    data = data.dropna().drop_duplicates(subset=colname)

data


# # Feature Engineering
# 
# Consider that the "bitter" taste is caused by the interaction of molecules with certain proteins (see [here](https://en.wikipedia.org/wiki/Taste_receptor#Bitter)). For a molecule to bind at the appropriate binding site on a protein, it will need to have an optimal structure, in terms of its shape, how it's polarized, whether it can form hydrogen bonds, etc. For this reason, we'll take a few different descriptors using RDKit.
# 
# - NuMAromaticRings, the number of aromatic rings in the moleclue. Aromatic rings are typically greasy, and generally add steric bulk to a molecule, so they'll likely affect binding to the protein.
# - NumHAcceptors, the number of hydrogen-bond acceptors (electronegative atoms with a lone pair) in the molecule. Proteins are littered with hydrogen bond donors/acceptors, so this will help provide information on the interaction with the protein.
# - NumHDonors, the number of hydogen-bond donors (hydrogens bound to an electronegative atom) in the molecule. See above justification for why this matters to the binding prediction.
# - MolWt, the molecular weight of the molecule. This is helpful as a description of the molecule's size, which in turn helps inform how it will bind to the protein.
# - MinPartialCharge, the minimum partial charge in the molecule, estimated by RDKit. This (and the MaxPartialCharge) are helpful, in that they will help give an idea of how polarized the molecule is. Proteins tend to have polar/nonpolar regions, so this will help inform where the molecule can bind to the protein.
# - MaxPartialCharge, the maximum partial charge in the molecule, estimated by RDKit. See above justification.
# - And finally, we'll take some various density metrics of the Morgan Fingerprints supplied by RDKit.
# 
# Additionally, some of the SMILES strings in the dataset are incorrectly formatted (e.g. missing a closing character, or some other problem). To get around this, we'll surround the calculation of each descriptor in a try/except block. If we fail to calculate a descriptor, we'll just drop the row entirely.

# In[]:


# Let's get a bunch of descriptors into the dataset
descriptors = ((Lipinski.NumAromaticRings, "NumAromaticRings"),
               (Lipinski.NumHAcceptors, "HBondAcceptors"),
               (Lipinski.NumHDonors, "HBondDonors"),
               (Descriptors.MolWt, "MolecularWeight"),
               (Descriptors.MinPartialCharge, "MinPartialCharge"),
               (Descriptors.MaxPartialCharge, "MaxPartialCharge"),
               (Descriptors.FpDensityMorgan1, "FPDensityMorgan1"),
               (Descriptors.FpDensityMorgan2, "FPDensityMorgan2"),
               (Descriptors.FpDensityMorgan3, "FPDensityMorgan3"),
              )

# The dataset is a little unclean, with some badly-formatted SMILEs strings. We'll skip errors there, and just return None instead.
def get_descriptor(smiles, fun):
    try:
        result = fun(MolFromSmiles(smiles))
    except:
        result = None
    return result

for descriptor, name in descriptors:
    data[name] = data["Canonical SMILES"].apply(functools.partial(get_descriptor, fun=descriptor))

# And finally, we'll drop rows that didn't make it through the featurization
data = data.dropna()


# # Train/Test Split
# 
# Next, we'll perform our train/test split. 10% of the data is kept out as a testing set, and 90% are kept in as a training set.
# 
# Also, we do have a slightly imbalanced dataset; roughly 2x as many non-bitter (0) compounds are in the dataset as there are bitter (1) compounds. To get around this, we'll use the SMOTE algorithm to appropriately upsample our dataset. Specifically, we'll use a variation of SMOTE that uses k-means to direct where it samples from the minority class, which a recent paper indicates outperforms vanilla SMOTE: https://arxiv.org/abs/1711.00837

# In[]:


print("Before Resample\n", data.Bitter.value_counts())

# Split into train/test sets
data_train, data_test = sklearn.model_selection.train_test_split(data, test_size=0.5)

x = data_train.drop(columns=["Name", "Canonical SMILES", "Bitter"]).to_numpy()
y = (data_train.Bitter).to_numpy()


# Upsample our training set with SMOTE
# Specifically, we'll use the K-Means variation of SMOTE - https://arxiv.org/abs/1711.00837
smote = imblearn.over_sampling.KMeansSMOTE(k_neighbors=10)
x_resample, y_resample = smote.fit_resample(x, y)


# # Prepare the Hyperparameter Optimization
# 
# Next, we'll use an sklearn pipeline with two steps:
# 1. A Standard Scaler, which standardizes each column such that it has a mean of 0 and standard deviation of 1. This is important, as many error metrics and algorithms operate on the assumption that (or, just work better when) the different descriptors are on similar scales.
# 2. XGBoost, a gradient boosting library which is a bit faster than the Sklearn implementation. In addition, it can apply an analogous method for dropout regularization to the trees, using the DART algorithm (see https://xgboost.readthedocs.io/en/latest/tutorials/dart.html)
# 
# We make a function that can create a pipeline in Sklearn, given a set of parameters. In Sklearn, we can access the parameters of an individual estimator by using the f"{estimator_name}\_\_{parameter}" syntax. The `create_pipeline` function takes in those parameters, and builds a pipeline.
# 
# Second, we have an `objective` function. The function tests a set of possible parameters for the pipeline, and returns a score (in this case the area under the ROC curve). Optuna's goal, then, is the find a set of hyperparameters that maximize the score of the training set. To help guard against overfitting, we use cross-validation to determine the score. Finally, we'll set Optuna to use hyperband optimization, which has been shown to perform well when combined with the default sampler - https://tech.preferred.jp/en/blog/how-we-implement-hyperband-in-optuna/

# In[]:


def create_pipeline(params):
    pipeline = sklearn.pipeline.Pipeline([
        
        ("StandardScaler", sklearn.preprocessing.StandardScaler()),
        ("XGBoostClassifier", xgboost.XGBClassifier( objective="binary:logistic",
                                                     eval_metric="logloss",
                                                     use_label_encoder=False))
    ])
    pipeline.set_params(**params)
    return pipeline
    
def objective(trial: optuna.Trial):
    with joblib.parallel_backend("loky"):      
        params = {
            "XGBoostClassifier__n_estimators": trial.suggest_int("XGBoostClassifier__n_estimators", 1, 1000),
            "XGBoostClassifier__eta": trial.suggest_loguniform("XGBoostClassifier__eta", 0.01, 0.4),
            "XGBoostClassifier__min_child_weight": trial.suggest_int("XGBoostClassifier__min_child_weight", 1,100),
            "XGBoostClassifier__max_depth": trial.suggest_int("XGBoostClassifier__max_depth", 3,50),
            "XGBoostClassifier__gamma": trial.suggest_uniform("XGBoostClassifier__gamma", 0, 1),
            "XGBoostClassifier__subsample": trial.suggest_uniform("XGBoostClassifier__subsample", 0, 1),
            "XGBoostClassifier__lambda": trial.suggest_loguniform("XGBoostClassifier__lambda", 0.1, 1000),
            "XGBoostClassifier__alpha": trial.suggest_loguniform("XGBoostClassifier__alpha", 0.1, 1000),
            "XGBoostClassifier__colsample_bytree": trial.suggest_uniform("XGBoostClassifier__colsample_bytree", 0.1,1),
        }
        pipeline = create_pipeline(params)
        
        score = sklearn.model_selection.cross_val_score(pipeline, x_resample, y_resample, scoring="neg_log_loss", n_jobs=1, cv=4).mean()
        return score

study = optuna.create_study(direction="maximize", pruner=optuna.pruners.HyperbandPruner(min_resource=10, max_resource='auto'))


# # Optimize the Hyperparameters
# 
# Optuna will then use Bayesian optimization to efficiently sample the hyperparameter search space. This is signiifcantly faster than the brute force method of testing all possible combinations of hyperparameters (Grid Search), and more-intelligently samples the hyperparameter space than a Random Search would.

# In[]:


study.optimize(objective, n_trials=128, n_jobs=16)


# # Fit the Model to the Entire Training Set
# Extract the best hyperparameters from the study and fit to our training set.
# 
# We'll also recall the cross-validated ROC AUC from the hyperparameter optimization.

# In[]:


pipeline = create_pipeline(study.best_params)
pipeline.fit(x,y)

# We don't need to re-evaluate the eror, since Optuna kept track of it for us.
cv_error = study.best_value

print(f"4-CV LogLoss: {-cv_error}")


# # Evaluate the model against the test set
# 
# Here, we'll take the model that was trained on the training set, and get the ROC AUC score for the test-set data.

# In[]:


test_preds = pipeline.predict(data_test.drop(columns=["Name", "Canonical SMILES", "Bitter"]).to_numpy())
test_probs = pipeline.predict_proba(data_test.drop(columns=["Name", "Canonical SMILES", "Bitter"]).to_numpy())
test_score = sklearn.metrics.roc_auc_score(y_true=data_test.Bitter, y_score=test_probs[:,1])
fpr, tpr, threshes = sklearn.metrics.roc_curve(y_true=data_test.Bitter, y_score=test_probs[:,1])
threshes[0] -= 1

print(f"Test ROC AUC: {test_score}")


# # Test Set ROC Curve
# 
# Next, we'll plot the ROC curve for the testing set, and we'll evaluate the AUC for it. AUC=1 is a a classifier that perfectly predicts the correct result, AUC=0 is a classifier that perfectly predicts the wrong result (e.g. correct predictions can be had by inverting its predictions), and AUC=0.5 is a classifier that is equally as good as random guessing.

# In[]:


# Plot the curve
def draw_roc_curve(fpr, tpr, threshes, title):
    fig, ax = plt.subplots(dpi=300)
    points = np.array([fpr, tpr]).T.reshape(-1, 1, 2)

    segments = np.concatenate([points[:-1], points[1:]], axis=1)
    norm = plt.Normalize(threshes.min(), threshes.max())

    lc = matplotlib.collections.LineCollection(segments, cmap='jet', norm=norm, linewidths=2)
    lc.set_array(threshes)
    line = ax.add_collection(lc)
    fig.colorbar(line, ax=ax).set_label('Threshold')

    # Padding to ensure we see the line
    ax.margins(0.01)

    # plt.plot(false_positive_rate, true_positive_rate, c=colors, label=f"ROC2 Cure, AUC={roc_auc}")
    plt.title(title)
    plt.xlabel("False Positive Rate")
    plt.ylabel("True Positive Rate")
    plt.show()
draw_roc_curve(fpr, tpr, threshes, f"Test Set ROC curve, AUC={np.round(test_score,3)}")


# # Fit the Model to the Entire Dataset
# 
# At this point, we've assessed that the cross-validated ROC AUC of the training set is good, and that the test-set ROC AUC doesn't indicate that we're overfitting by much. We can now train the model on the entire dataset, using cross-validation to get an error estimate of the model.

# In[]:


# Now that we have a good idea of our error, let's retrain on the entire dataset
x_full = data.drop(columns=["Name", "Canonical SMILES", "Bitter"]).to_numpy()
y_full = data.Bitter.to_numpy()

# First, upsample using smote, as we did with the training set
x_full_resample, y_full_resample = smote.fit_resample(x_full, y_full)

# Finally, fit to the entire dataset
pipeline.fit(x_full_resample,y_full_resample)


# # ROC Curve - Entire Dataset
# 
# Finally, we'll take an ROC curve for when we use the entire dataset as the training set.

# In[]:


final_probs = pipeline.predict_proba(x_full)
final_score = sklearn.metrics.roc_auc_score(y_true=y_full, y_score=final_probs[:,1])
fpr, tpr, threshes = sklearn.metrics.roc_curve(y_true=y_full, y_score=final_probs[:,1])
threshes[0] -= 1
draw_roc_curve(fpr, tpr, threshes, f"Full Dataset Set ROC curve, AUC={np.round(final_score,3)}")


# # Checking Specific Cases
# As a sanity check, we'll look at the cases of Denatonium and Glycerine.
# 
# Denatonium is the most-bitter chemical known; it's oftentimes used as a bittering agent.
# 
# Glycerine (aka Glycerol) is a relatively sweet substance, used to make a variety of products. For example, its nitration leads to the production of nitroglycerine.
# 

# In[]:


# Denatonium! Very bitter. It doesn't seem to be in the dataset.
print(data[data.Name == "Denatonium"])
print(data[data.Name == "Bitrex"])

# Glycerol! Probably tastes kinda sweet, but most certainly not bitter.
print(data[data.Name == "Glycerol"])


# In[]:


names = [
    "Denatonium", 
    "Ethanol"
]
smiles = [
    "[O-]C(=O)c1ccccc1.Cc2cccc(C)c2NC(=O)C[N+](CC)(CC)Cc3ccccc3",
    "OCC"
]

new_data = pd.DataFrame.from_dict({
    "Name": names,
    "SMILES": smiles
})

for descriptor, name in descriptors: # Turn into a function
    new_data[name] = new_data.SMILES.apply(functools.partial(get_descriptor, fun=descriptor))
    
new_data["Predictions"] = list(map(lambda pred: "Bitter" if pred else "Non-Bitter", pipeline.predict(new_data.drop(columns=["Name", "SMILES"]))))
new_data


# # Interpreting the Model: Feature Importance
# 
# Recall that the whether a molecule is bitter corresponds to whether it will bind strongly to a protein.
# 
# Without investigating the binding site of the protein itself (which would be a good idea if we wanted to study this further), we can check empirically whether a particular structural feature is important to whether a molecule will bind to the protein or not. This can give hints about how we might design a molecule that binds (or doesn't bind) to the receptors involved in determining bitterness.

# In[]:


importances = pipeline["XGBoostClassifier"].feature_importances_
column_names = data_train.drop(columns=["Name", "Canonical SMILES", "Bitter"]).columns

plt.barh(range(len(importances)), sorted(importances))
plt.gca().set_yticks(range(len(importances)))
plt.gca().set_yticklabels(column_names[importances.argsort()])

plt.show()


# In[ ]:







"""TITLE: Place a 2D material Layer on a Surface."""

"""BLOCK: Install Dependencies"""
"""
This scripts creates an interface between two materials: (1) substrate and (2) layer.
It is assumed that the substrate is a 3D material and the layer is a 2D material.
This script is based on the ASE package (Atomic Simulation Environment) and the tutorial at:
https://wiki.fysik.dtu.dk/ase/gettingstarted/manipulating_atoms/manipulating_atoms.html#interface-building.
"""

# This block handles the installation of necessary packages for the script using Micropip in a Pyodide environment.
# ASE (Atomic Simulation Environment) is installed for manipulating structures.
import micropip

await micropip.install("ase")
print("installed ase")

"""BLOCK: Utils, Class Definitions, and main()"""
"""
NOTE: edit the variables below for your specific use case.
"""

# Indices identify the substrate and layer from the list of input materials under `materials_in` in globals().
SUBSTRATE_INDEX = 0
LAYER_INDEX = 1

SETTINGS = {
    "substrate_surface": {
        # Set Miller indices as a tuple for the resulting substrate surface.
        "miller_indices": (1, 1, 1),
        # The vacuum space (in Ångströms) added to the surface in the direction perpendicular to the surface.
        "vacuum": 5,
        # The number of atomic layers in the resulting substrate.
        "number_of_layers": 3,
        # The transformation matrix for the surface. Format is: [[v1x, v1y], [v2x, v2y]].
        # fmt: off
        "superlattice_matrix": [
            [1, 0],
            [0, 1]
        ],
        # fmt: on
    },
    "layer_surface": {
        # Set Miller indices as a tuple for the resulting layer surface: (0,0,1) for 2D material
        "miller_indices": (0, 0, 1),
        # The vacuum space (in Ångströms) added to the surface in the direction perpendicular to the surface.
        "vacuum": 5,
        # The number of atomic layers in the resulting substrate: 1 for 2D material
        "number_of_layers": 1,
        # The transformation matrix for the surface. Format is: [[v1x, v1y], [v2x, v2y]].
        # fmt: off
        "superlattice_matrix": [
            [1, 0],
            [0, 1]
        ],
        # fmt: on
    },
    "interface": {
        "distance": 3.0,
    },
    # If True the layer cell and basis vectors will be scaled to fit the substrate cell.
    # Mind the strain that is introduced by this operation.
    "scale_layer_to_fit": False,
}


"""
NOTE: DO NOT edit code below unless you know what you are doing.
"""

from ase.build import surface as make_surface, supercells
from ase.io import read, write
import io
import numpy as np
import io
from ase.build import surface, make_supercell
from ase.io import read, write
import numpy as np


# Utility functions
def poscar_to_atoms(poscar):
    input = io.StringIO(poscar)
    atoms = read(input, format="vasp")

    return atoms


def atoms_to_poscar(atoms):
    output = io.StringIO()
    write(output, atoms, format="vasp")
    content = output.getvalue()
    output.close()

    return content


def expand_matrix_2x2_to_3x3(matrix_2x2):
    matrix_3x3 = np.identity(3)
    matrix_3x3[0:2, 0:2] = matrix_2x2

    return matrix_3x3


def create_surface_and_supercell(atoms, miller_indices, number_of_layers, vacuum, superlattice_matrix):
    """
    Creates a surface and supercell based on the input parameters.

    Params:
        atoms: The input atoms.
        miller_indices: The Miller indices for the surface.
        number_of_layers: The number of layers in the resulting surface.
        vacuum: The vacuum space (in Ångströms) added to the surface in the direction perpendicular to the surface.
        superlattice_matrix: The transformation matrix for the surface. Format is: [[v1x, v1y], [v2x, v2y]].

    Returns:
        supercell_atoms: The resulting supercell.
    """

    surface_atoms = surface(atoms, miller_indices, number_of_layers, vacuum)
    expanded_matrix = expand_matrix_2x2_to_3x3(superlattice_matrix)
    supercell_atoms = make_supercell(surface_atoms, expanded_matrix)

    return supercell_atoms


def calculate_strain_matrix(scaled_layer_cell, original_layer_cell):
    """
    Calculates the strain matrix based on the scaled and original layer cells.

    Params:
        scaled_layer_cell: The scaled layer cell.
        original_layer_cell: The original layer cell.

    Returns:
        difference_matrix: The difference matrix.
    """

    cell1 = np.array(original_layer_cell)[0:2, 0:2]
    cell2 = np.array(scaled_layer_cell)[0:2, 0:2]
    transformation_matrix = np.linalg.inv(cell1) @ cell2
    difference_matrix = transformation_matrix - np.eye(2)

    return difference_matrix


def create_interface(
    substrate_poscar,
    layer_poscar,
    substrate_surface_settings,
    layer_surface_settings,
    interface_settings,
    scale_layer_to_fit=False,
):
    """
    Creates an interface between the substrate and the layer.

    Params:
        substrate_poscar: The POSCAR data for the substrate.
        layer_poscar: The POSCAR data for the layer.
        substrate_surface_settings: The settings for the substrate surface.
        layer_surface_settings: The settings for the layer surface.
        interface_settings: The settings for the interface.
        scale_layer_to_fit: If True the layer cell and basis vectors will be scaled to fit the substrate cell.

    Returns:
        interface: The resulting interface.
    """

    substrate_atoms = poscar_to_atoms(substrate_poscar)
    layer_atoms = poscar_to_atoms(layer_poscar)

    substrate_supercell = create_surface_and_supercell(
        substrate_atoms,
        miller_indices=substrate_surface_settings["miller_indices"],
        number_of_layers=substrate_surface_settings["number_of_layers"],
        vacuum=substrate_surface_settings["vacuum"],
        superlattice_matrix=substrate_surface_settings["superlattice_matrix"],
    )

    layer_supercell = create_surface_and_supercell(
        layer_atoms,
        miller_indices=layer_surface_settings["miller_indices"],
        number_of_layers=layer_surface_settings["number_of_layers"],
        vacuum=layer_surface_settings["vacuum"],
        superlattice_matrix=layer_surface_settings["superlattice_matrix"],
    )

    # Calculate strain on the layer
    m = calculate_strain_matrix(substrate_supercell.get_cell(), layer_supercell.get_cell())
    m_percent = m * 100
    percent_str = np.array2string(m_percent, formatter={"float": "{:0.2f}%".format})
    print("Strain matrix as percentages:\n", percent_str)

    # Scale the layer to fit the substrate (mind the strain)
    if scale_layer_to_fit:
        layer_supercell.set_cell(substrate_supercell.get_cell(), scale_atoms=True)
        layer_supercell.wrap()

    # Adjust Z position based on the Z offset
    z_max_substrate = max(substrate_supercell.positions[:, 2])
    z_min_layer = min(layer_supercell.positions[:, 2])
    z_offset = z_max_substrate - z_min_layer + interface_settings["distance"]
    layer_supercell.positions[:, 2] += z_offset

    # Combine substrate and layer into one Atoms object
    interface = substrate_supercell + layer_supercell
    return interface


def main():
    """
    The main function of the script.
    Creates an interface between the substrate and the layer based on the SETTINGS defined above.
    Makes the resulting material available to the platform JS environment.

    Params:
        (implicit) globals(): The data passed from the platform JS environment.
            - materials_in: The list of input materials. Materials are represented as JS classes. See
                https://github.com/Exabyte-io/made.js/blob/dev/src/material.js#L293 for more information..

    Returns:
        globals(): The globals() dictionary is returned to the platform JS environment.
            - materials_out: The list of output materials with poscar as strucutre representation.
    """
    # Get the input materials from the platform JS environment.
    materials = globals()["materials_in"]

    # Get poscar data for the substrate and the layer.
    # Per https://github.com/Exabyte-io/made.js/blob/dev/src/material.js#L293
    substrate_data = materials[SUBSTRATE_INDEX].getAsPOSCAR()
    layer_data = materials[LAYER_INDEX].getAsPOSCAR()

    # Interface is created based on SETTINGS for both substrate and layer
    interface_structure = create_interface(
        substrate_poscar=substrate_data,
        layer_poscar=layer_data,
        substrate_surface_settings=SETTINGS["substrate_surface"],
        layer_surface_settings=SETTINGS["layer_surface"],
        interface_settings=SETTINGS["interface"],
        scale_layer_to_fit=SETTINGS["scale_layer_to_fit"],
    )

    # Make the data available to the platform JS environment.
    globals()["materials_out"] = [
        {
            "poscar": atoms_to_poscar(interface_structure),
        }
    ]
    return globals()


main()



#!/usr/bin/env python
# coding: utf-8

# <a href="https://colab.research.google.com/github/Exabyte-io/api-examples/blob/dev/examples/material/api_interoperability_showcase.ipynb" target="_parent">
# <img alt="Open in Google Colab" src="https://user-images.githubusercontent.com/20477508/128780728-491fea90-9b23-495f-a091-11681150db37.jpeg" width="150" border="0">
# </a>

# # Overview
# 
# This example was created as part of our [Advanced Topics Webinar](https://www.youtube.com/watch?v=psSFC409jSg) on February 19, 2021. This webinar focused on explaining our API in detail, and provided examples of many areas of its functionality.
# 
# In this notebook, we showcase a major advantage of APIs: interoperability. We begin by performing a query using the [Materials Project](https://materialsproject.org) API for all systems containing Iron and Oxygen. We then filter our results (for demonstraiton purposes, we keep only the first 10 materials found). Finally, we upload our results to the Mat3ra platform, where further calculations could be performed to characterize these materials.

# # Complete Authorization Form and Initialize Settings
# 
# This will also determine environment and set all environment variables. We determine if we are using Jupyter Notebooks or Google Colab to run this tutorial.
# 
# If you are running this notebook from Google Colab, Colab takes ~1 min to execute the following cell.
# 
# ACCOUNT_ID and AUTH_TOKEN - Authentication parameters needed for when making requests to [Mat3ra.com's API Endpoints](https://docs.mat3ra.com/rest-api/endpoints/).
# 
# MATERIALS_PROJECT_API_KEY - Authentication parameter needed for when making requests to [Material Project's API](https://materialsproject.org/open)
# 
# ORGANIZATION_ID - Authentication parameter needed for when working with collaborative accounts https://docs.mat3ra.com/collaboration/organizations/overview/
# 
# > <span style="color: orange">**NOTE**</span>: If you are running this notebook from Jupyter, the variables ACCOUNT_ID, AUTH_TOKEN, MATERIALS_PROJECT_API_KEY, and ORGANIZATION_ID should be set in the file [settings.json](../../utils/settings.json) if you need to use these variables. To obtain API token parameters, please see the following link to the documentation explaining how to get them: https://docs.mat3ra.com/accounts/ui/preferences/api/

# In[ ]:


# @title Authorization Form
ACCOUNT_ID = "ACCOUNT_ID"  # @param {type:"string"}
AUTH_TOKEN = "AUTH_TOKEN"  # @param {type:"string"}
MATERIALS_PROJECT_API_KEY = "MATERIALS_PROJECT_API_KEY"  # @param {type:"string"}
ORGANIZATION_ID = "ORGANIZATION_ID"  # @param {type:"string"}

import os

if "COLAB_JUPYTER_IP" in os.environ:
    os.environ.update(
        dict(
            ACCOUNT_ID=ACCOUNT_ID,
            AUTH_TOKEN=AUTH_TOKEN,
            MATERIALS_PROJECT_API_KEY=MATERIALS_PROJECT_API_KEY,
            ORGANIZATION_ID=ORGANIZATION_ID,
        )
    )

    get_ipython().system('GIT_BRANCH="dev"; export GIT_BRANCH; curl -s "https://raw.githubusercontent.com/Exabyte-io/api-examples/${GIT_BRANCH}/scripts/env.sh" | bash')


# # Imports

# In[ ]:


from utils.settings import ENDPOINT_ARGS, MATERIALS_PROJECT_API_KEY
from utils.generic import display_JSON

import ase.io
from pymatgen.ext.matproj import MPRester

from exabyte_api_client.endpoints.materials import MaterialEndpoints


# # Query the Materials Project
# 
# We begin by using the Materials Project API implemented in [PyMatGen](https://pymatgen.org/pymatgen.ext.matproj.html) to perform a query for all systems containing Iron and Oxygen.

# # Query Materials Project for all systems containing Iron and Oxygen

# In[ ]:


materials_project_api = MPRester(MATERIALS_PROJECT_API_KEY)
iron_oxides_ids = materials_project_api.get_materials_ids("Fe-O")

print(iron_oxides_ids)


# # Filtering the Results
# 
# This returns a lot of materials - 160 to be exact! In many cases, it is useful to filter down the number of materials. For example, we may want to exclude large unit cells that may be computationally intensive to study. Or we may want to restrict our results to only thermodynamically-stable oxides, by use of the material's [energy above hull](https://wiki.materialsproject.org/Glossary_of_Terms).
# 
# As a basic example, here we only keep the first 10 iron oxides that the Materials Project API returned to us, and discard the other 150.

# In[ ]:


# As a demonstration, take the first 10 iron oxides
some_iron_oxides = iron_oxides_ids[:10]
print(some_iron_oxides)


# # Bringing Materials Into the User Account
# 
# Now that we have filtered the results from Materials Project down to just 10 structures, we may want to study them further with the computational models provided by Mat3ra. For example, we may be interested in leveraging a DFT code to find the structure with the largest band-gap, or perhaps we want to conduct a high-throughput screening of each material's surface energies.

# In[ ]:


# Upload the first 10 iron oxides found to our account
exabyte_materials_api = MaterialEndpoints(*ENDPOINT_ARGS)
materials = exabyte_materials_api.import_from_materialsproject(MATERIALS_PROJECT_API_KEY, some_iron_oxides)


# Finally, it is always useful to stay organized. Materials sets make this convenient, acting as a folder to keep a group of related materials in. This would be especially helpful if, in the future, we wanted run a calculation over all the oxides we found in this example.

# In[ ]:


# Move the iron oxides to a materials set, just for this example
materials_set = exabyte_materials_api.create_set({"name": "Some Iron Oxides"})
for material in materials:
    exabyte_materials_api.move_to_set(material["_id"], "", materials_set["_id"])




"""TITLE: Custom Transformation (Example)."""
"""BLOCK: Main"""
"""
This code demonstrates how to:
 - get materials from the application runtime as JS class instances, 
 - extract poscar data from the materials, 
 - return the data back to the application runtime.
See https://github.com/Exabyte-io/made.js/blob/dev/src/material.js also.
"""
import json

# materials_in holds the array of materials pased from the application runtime
input_materials = materials_in


def main():
    material = input_materials[0]
    poscar = material.getAsPOSCAR()

    # REMOVE THIS COMMENT AND PLACE YOUR CODE HERE.
    # NOTE: when passing back to the application runtime, the poscar data will be converted to a JS class instance
    # of a Material class (https://github.com/Exabyte-io/made.js/blob/dev/src/material.js).

    output_materials = [{"poscar": poscar}]
    print("Output materials: ", json.dumps(output_materials, indent=4))

    # Return the globals() dictionary to the application environment. Required for proper operation at this time.
    globals()["materials_out"] = output_materials
    return globals()


main()



"""TITLE: Strain Matching of a 2D Material Layer on a Surface."""

"""BLOCK: Install Dependencies"""

"""
This scripts installs the required packages for a basic Pyodide environment for Materials structure manipulation.
It uses micropip - a Python package manager for Pyodide.
"""
import micropip

packages = [
    "https://files.mat3ra.com:44318/uploads/pymatgen-2023.9.10-py3-none-any.whl",
    "https://files.mat3ra.com:44318/web/pyodide/spglib-2.0.2-py3-none-any.whl",
    "https://files.pythonhosted.org/packages/d9/0e/2a05efa11ea33513fbdf4a2e2576fe94fd8fa5ad226dbb9c660886390974/ruamel.yaml-0.17.32-py3-none-any.whl",
    "ase==3.22.1",
    "networkx==3.2.1",
    "monty==2023.11.3",
    "scipy==1.11.1",
    "lzma",
    "tabulate==0.9.0",
    "sqlite3",
    "sympy==1.12",
    "uncertainties==3.1.6",
]


async def install_package(pkg):
    """
    Installs a package in a Pyodide environment.
    Args:
        pkg: The name of the package to install.

    Returns:
        None
    """
    is_url = pkg.startswith("http://") or pkg.startswith("https://")
    are_dependencies_installed = not is_url
    await micropip.install(pkg, deps=are_dependencies_installed)
    # Extract package name for printing
    pkg_name = pkg.split("/")[-1].split("-")[0] if is_url else pkg.split("==")[0]
    print(f"Installed {pkg_name}")


for package in packages:
    await install_package(package)

"""BLOCK: Utils, Class Definitions, and main()"""
"""
This script employs the Zur and McGill SuperLattice algorithm (Journal of Applied Physics 55 (1984), 378 ; doi: 10.1063/1.333084)  for strain matching a 2D material layer on a surface. 
Using pymatgen, it constructs coherent interfaces between a substrate and a film layer, with a focus on lattice matching and interface terminations. 
Link to pymatgen ZSL implementation: https://pymatgen.org/pymatgen.analysis.interfaces.html#pymatgen.analysis.interfaces.zsl.ZSLGenerator
Key parameters like Miller indices and layer thicknesses can be customized.
The result of the lagorithm is a list of interfaces, sorted by the mean absolute strain.
Plot shows the mean absolute strain vs. the number of atoms in the interface to select corresponding material by index. 
"""
from __future__ import annotations


# Select materials from the list of input materials under `materials_in` in globals().
SUBSTRATE_INDEX = 0
LAYER_INDEX = 1

# Select interfaces from list of generated interfaces sorted by increasing mean absolute strain
# By interval or by index, e.g. [0, 10] or 0
OUTPUT_INDICES = [0, 10]

# Miller indices and thickness for substrate and layer.
SUBSTRATE_MILLER = (1, 1, 1)
SUBSTRATE_THICKNESS = 3
LAYER_MILLER = (0, 0, 1)
LAYER_THICKNESS = 1

# Distance between layers, in Angstroms
DISTANCE = 3.0

# Parameters for the ZSL algorithm
# as defined in: https://pymatgen.org/pymatgen.analysis.interfaces.html#pymatgen.analysis.interfaces.zsl.ZSLGenerator
MAX_AREA = 400
MAX_AREA_TOL = 0.09
MAX_LENGTH_TOL = 0.03
MAX_ANGLE_TOL = 0.01

# Strains within this tolerance are considered equal
STRAIN_TOL = 10e-6

# Plot settings
# Strain axis limits in percent
X_MIN = 0.01
X_MAX = 100

# Number of atoms axis limits
Y_MIN = 1
Y_MAX = 1000


"""
NOTE: DO NOT edit code below unless you know what you are doing.
"""

""" 
Classes and Definitions
"""

from itertools import product
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
from numpy.testing import assert_allclose
from scipy.linalg import polar

import pymatgen
from pymatgen.analysis.interfaces.zsl import ZSLGenerator as PymatgenZSL
from pymatgen.core.structure import Structure
from operator import itemgetter
from pymatgen.analysis.elasticity.strain import Deformation
from pymatgen.analysis.interfaces.zsl import ZSLGenerator, fast_norm
from pymatgen.core.interface import Interface, label_termination
from pymatgen.core.surface import SlabGenerator

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

    from pymatgen.core import Structure


class CoherentInterfaceBuilder:
    """
    This class constructs the coherent interfaces between two crystalline slabs
    Coherency is defined by matching lattices not sub-planes.
    """

    def __init__(
        self,
        substrate_structure: Structure,
        film_structure: Structure,
        film_miller: tuple[int, int, int],
        substrate_miller: tuple[int, int, int],
        zslgen: ZSLGenerator | None = None,
    ):
        """
        Args:
            substrate_structure: structure of substrate
            film_structure: structure of film
            film_miller: miller index of the film layer
            substrate_miller: miller index for the substrate layer
            zslgen: BiDirectionalZSL if you want custom lattice matching tolerances for coherency.
        """
        # Bulk structures
        self.substrate_structure = substrate_structure
        self.film_structure = film_structure
        self.film_miller = film_miller
        self.substrate_miller = substrate_miller
        self.zslgen = zslgen or ZSLGenerator(bidirectional=True)

        self._find_matches()
        self._find_terminations()

    def _find_matches(self) -> None:
        """
        Finds and stores the ZSL matches.
        """
        self.zsl_matches = []

        film_sg = SlabGenerator(
            self.film_structure,
            self.film_miller,
            min_slab_size=1,
            min_vacuum_size=3,
            in_unit_planes=True,
            center_slab=True,
            primitive=True,
            reorient_lattice=False,  # This is necessary to not screw up the lattice
        )

        sub_sg = SlabGenerator(
            self.substrate_structure,
            self.substrate_miller,
            min_slab_size=1,
            min_vacuum_size=3,
            in_unit_planes=True,
            center_slab=True,
            primitive=True,
            reorient_lattice=False,  # This is necessary to not screw up the lattice
        )

        film_slab = film_sg.get_slab(shift=0)
        sub_slab = sub_sg.get_slab(shift=0)

        film_vectors = film_slab.lattice.matrix
        substrate_vectors = sub_slab.lattice.matrix

        # Generate all possible interface matches
        self.zsl_matches = list(self.zslgen(film_vectors[:2], substrate_vectors[:2], lowest=False))

        for match in self.zsl_matches:
            xform = get_2d_transform(film_vectors, match.film_vectors)
            strain, rot = polar(xform)
            assert_allclose(
                strain, np.round(strain), atol=1e-12
            ), "Film lattice vectors changed during ZSL match, check your ZSL Generator parameters"

            xform = get_2d_transform(substrate_vectors, match.substrate_vectors)
            strain, rot = polar(xform)
            assert_allclose(
                strain, strain.astype(int), atol=1e-12
            ), "Substrate lattice vectors changed during ZSL match, check your ZSL Generator parameters"

    def _find_terminations(self):
        """
        Finds all terminations.
        """
        film_sg = SlabGenerator(
            self.film_structure,
            self.film_miller,
            min_slab_size=1,
            min_vacuum_size=3,
            in_unit_planes=True,
            center_slab=True,
            primitive=True,
            reorient_lattice=False,  # This is necessary to not screw up the lattice
        )

        sub_sg = SlabGenerator(
            self.substrate_structure,
            self.substrate_miller,
            min_slab_size=1,
            min_vacuum_size=3,
            in_unit_planes=True,
            center_slab=True,
            primitive=True,
            reorient_lattice=False,  # This is necessary to not screw up the lattice
        )

        film_slabs = film_sg.get_slabs()
        sub_slabs = sub_sg.get_slabs()

        film_shits = [s.shift for s in film_slabs]
        film_terminations = [label_termination(s) for s in film_slabs]

        sub_shifts = [s.shift for s in sub_slabs]
        sub_terminations = [label_termination(s) for s in sub_slabs]

        self._terminations = {
            (film_label, sub_label): (film_shift, sub_shift)
            for (film_label, film_shift), (sub_label, sub_shift) in product(
                zip(film_terminations, film_shits), zip(sub_terminations, sub_shifts)
            )
        }
        self.terminations = list(self._terminations)

    def get_interfaces(
        self,
        termination: tuple[str, str],
        gap: float = 2.0,
        vacuum_over_film: float = 20.0,
        film_thickness: float = 1,
        substrate_thickness: float = 1,
        in_layers: bool = True,
    ) -> Iterator[Interface]:
        """
        Generates interface structures given the film and substrate structure
        as well as the desired terminations.

        Args:
            termination (tuple[str, str]): termination from self.termination list
            gap (float, optional): gap between film and substrate. Defaults to 2.0.
            vacuum_over_film (float, optional): vacuum over the top of the film. Defaults to 20.0.
            film_thickness (float, optional): the film thickness. Defaults to 1.
            substrate_thickness (float, optional): substrate thickness. Defaults to 1.
            in_layers (bool, optional): set the thickness in layer units. Defaults to True.

        Yields:
            Iterator[Interface]: interfaces from slabs
        """
        film_sg = SlabGenerator(
            self.film_structure,
            self.film_miller,
            min_slab_size=film_thickness,
            min_vacuum_size=3,
            in_unit_planes=in_layers,
            center_slab=True,
            primitive=True,
            reorient_lattice=False,  # This is necessary to not screw up the lattice
        )

        sub_sg = SlabGenerator(
            self.substrate_structure,
            self.substrate_miller,
            min_slab_size=substrate_thickness,
            min_vacuum_size=3,
            in_unit_planes=in_layers,
            center_slab=True,
            primitive=True,
            reorient_lattice=False,  # This is necessary to not screw up the lattice
        )

        film_shift, sub_shift = self._terminations[termination]

        film_slab = film_sg.get_slab(shift=film_shift)
        sub_slab = sub_sg.get_slab(shift=sub_shift)

        for match in self.zsl_matches:
            # Build film superlattice
            super_film_transform = np.round(
                from_2d_to_3d(get_2d_transform(film_slab.lattice.matrix[:2], match.film_sl_vectors))
            ).astype(int)
            film_sl_slab = film_slab.copy()
            film_sl_slab.make_supercell(super_film_transform)
            assert_allclose(
                film_sl_slab.lattice.matrix[2], film_slab.lattice.matrix[2], atol=1e-08
            ), "2D transformation affected C-axis for Film transformation"
            assert_allclose(
                film_sl_slab.lattice.matrix[:2], match.film_sl_vectors, atol=1e-08
            ), "Transformation didn't make proper supercell for film"

            # Build substrate superlattice
            super_sub_transform = np.round(
                from_2d_to_3d(get_2d_transform(sub_slab.lattice.matrix[:2], match.substrate_sl_vectors))
            ).astype(int)
            sub_sl_slab = sub_slab.copy()
            sub_sl_slab.make_supercell(super_sub_transform)
            assert_allclose(
                sub_sl_slab.lattice.matrix[2], sub_slab.lattice.matrix[2], atol=1e-08
            ), "2D transformation affected C-axis for Film transformation"
            assert_allclose(
                sub_sl_slab.lattice.matrix[:2], match.substrate_sl_vectors, atol=1e-08
            ), "Transformation didn't make proper supercell for substrate"

            # Add extra info
            match_dict = match.as_dict()
            interface_properties = {k: match_dict[k] for k in match_dict if not k.startswith("@")}

            dfm = Deformation(match.match_transformation)

            strain = dfm.green_lagrange_strain
            interface_properties["strain"] = strain
            interface_properties["von_mises_strain"] = strain.von_mises_strain
            interface_properties["termination"] = termination
            interface_properties["film_thickness"] = film_thickness
            interface_properties["substrate_thickness"] = substrate_thickness

            yield {
                "interface": Interface.from_slabs(
                    substrate_slab=sub_sl_slab,
                    film_slab=film_sl_slab,
                    gap=gap,
                    vacuum_over_film=vacuum_over_film,
                    interface_properties=interface_properties,
                ),
                "strain": strain,
                "von_mises_strain": strain.von_mises_strain,
                "mean_abs_strain": round(np.mean(np.abs(strain)) / STRAIN_TOL) * STRAIN_TOL,
                "film_sl_vectors": match.film_sl_vectors,
                "substrate_sl_vectors": match.substrate_sl_vectors,
                "film_transform": super_film_transform,
                "substrate_transform": super_sub_transform,
            }


def get_rot_3d_for_2d(film_matrix, sub_matrix) -> np.ndarray:
    """
    Find transformation matrix that will rotate and strain the film to the substrate while preserving the c-axis.
    """
    film_matrix = np.array(film_matrix)
    film_matrix = film_matrix.tolist()[:2]
    film_matrix.append(np.cross(film_matrix[0], film_matrix[1]))

    # Generate 3D lattice vectors for substrate super lattice
    # Out of plane substrate super lattice has to be same length as
    # Film out of plane vector to ensure no extra deformation in that
    # direction
    sub_matrix = np.array(sub_matrix)
    sub_matrix = sub_matrix.tolist()[:2]
    temp_sub = np.cross(sub_matrix[0], sub_matrix[1]).astype(float)  # conversion to float necessary if using numba
    temp_sub = temp_sub * fast_norm(
        np.array(film_matrix[2], dtype=float)
    )  # conversion to float necessary if using numba
    sub_matrix.append(temp_sub)

    transform_matrix = np.transpose(np.linalg.solve(film_matrix, sub_matrix))

    rot, _ = polar(transform_matrix)

    return rot


def get_2d_transform(start: Sequence, end: Sequence) -> np.ndarray:
    """
    Gets a 2d transformation matrix
    that converts start to end.
    """
    return np.dot(end, np.linalg.pinv(start))


def from_2d_to_3d(mat: np.ndarray) -> np.ndarray:
    """
    Converts a 2D matrix to a 3D matrix.
    """
    new_mat = np.diag([1.0, 1.0, 1.0])
    new_mat[:2, :2] = mat
    return new_mat


def plot_strain_vs_atoms(strain_mode, sorted_interfaces):
    """
    Plots the strain vs. the number of atoms in the interface. With hover-over labels.
    """
    fig, ax = plt.subplots()

    # Scatter plot
    x = [i[strain_mode] * 100 for i in sorted_interfaces]  # in precentage
    y = [i["interface"].num_sites for i in sorted_interfaces]
    sc = ax.scatter(x, y)

    # Annotation for the hover-over labels
    annot = ax.annotate(
        "",
        xy=(0, 0),
        xytext=(20, 20),
        textcoords="offset points",
        bbox=dict(boxstyle="round", fc="w"),
        arrowprops=dict(arrowstyle="->"),
    )
    annot.set_visible(False)

    def update_annot(ind):
        pos = sc.get_offsets()[ind["ind"][0]]
        annot.xy = pos
        text = "{}".format(" ".join([str(index) for index in ind["ind"]]))
        annot.set_text(text)
        annot.get_bbox_patch().set_alpha(0.4)

    def hover(event):
        vis = annot.get_visible()
        if event.inaxes == ax:
            cont, ind = sc.contains(event)
            if cont:
                update_annot(ind)
                annot.set_visible(True)
                fig.canvas.draw_idle()
            else:
                if vis:
                    annot.set_visible(False)
                    fig.canvas.draw_idle()

    # Connect the hover event
    fig.canvas.mpl_connect("motion_notify_event", hover)

    # Set the scale and labels
    plt.xscale("log")
    plt.yscale("log")
    plt.xlim(X_MIN, X_MAX)
    plt.ylim(Y_MIN, Y_MAX)
    plt.xlabel("strain in %")
    plt.ylabel("number of atoms")

    plt.show()


def output_materials(sorted_interfaces, output_indices, strain_mode):
    """
    Outputs the materials in the output range.
    """
    # Initialize an empty list to store the materials
    materials_out = []

    # Check if output_indices is a single number or a list
    if isinstance(output_indices, int):
        # If it's a single number, output the material with that index
        output_range = [output_indices]
    else:
        # If it's a list, output materials in that range
        output_range = range(*output_indices)

    # Loop over the interfaces in the output range
    for index in output_range:
        best_interface = sorted_interfaces[index]["interface"]

        wrapped_structure = best_interface.copy()
        # This should wrap the atoms inside the unit cell
        wrapped_structure.make_supercell([1, 1, 1])
        interface_poscar = wrapped_structure.to(fmt="poscar")

        # Add the strain to the POSCAR first line to use as a name
        strain = sorted_interfaces[index][strain_mode]
        lines = interface_poscar.split("\n")
        lines[0] = lines[0] + " strain: " + "{:.3f}".format(strain * 100) + "%"
        interface_poscar = "\n".join(lines)

        # Append the material to the list
        materials_out.append({"poscar": interface_poscar})

    return materials_out


""" MAIN """


def main():
    # Interaction with Platform
    materials_in = globals()["materials_in"]
    substrate = Structure.from_str(materials_in[SUBSTRATE_INDEX].getAsPOSCAR(), fmt="poscar")
    layer = Structure.from_str(materials_in[LAYER_INDEX].getAsPOSCAR(), fmt="poscar")

    # Create Interface Builder class
    zsl = PymatgenZSL(
        max_area_ratio_tol=MAX_AREA_TOL, max_area=MAX_AREA, max_length_tol=MAX_LENGTH_TOL, max_angle_tol=MAX_ANGLE_TOL
    )
    cib = CoherentInterfaceBuilder(
        substrate_structure=substrate,
        film_structure=layer,
        substrate_miller=SUBSTRATE_MILLER,
        film_miller=LAYER_MILLER,
        zslgen=zsl,
    )

    # Run the Interface Building process
    cib._find_terminations()
    matches = cib.zsl_matches
    terminations = cib.terminations

    # Create interfaces
    interfaces = []
    for termination in terminations:
        interfaces = list(
            cib.get_interfaces(
                termination,
                gap=DISTANCE,
                film_thickness=LAYER_THICKNESS,
                substrate_thickness=SUBSTRATE_THICKNESS,
                in_layers=True,
            )
        )

    print(f"Found {len(matches)} interfaces")
    print(f"Found {len(terminations)} terminations:", terminations)

    strain_modes = {"VON_MISES": "von_mises_strain", "STRAIN": "strain", "MEAN": "mean_abs_strain"}
    strain_mode = strain_modes["MEAN"]
    interfaces_list = list(interfaces)

    # Sort interfaces by ascending strain and then by ascending number of atoms
    sorted_interfaces = sorted(interfaces_list, key=lambda x: (itemgetter(strain_mode)(x), x["interface"].num_sites))

    print("Interface with lowest strain (index 0):")
    print("    strain:", sorted_interfaces[0][strain_mode] * 100, "%")
    print("    number of atoms:", sorted_interfaces[0]["interface"].num_sites)

    # plot stran vs number of atoms via matplotlib
    plot_strain_vs_atoms(strain_mode, sorted_interfaces)

    # Return created materials to the platform
    globals()["materials_out"] = output_materials(sorted_interfaces, OUTPUT_INDICES, strain_mode)

    return globals()


main()



#!/usr/bin/env python
# coding: utf-8

# # Overview
# 
# This example has been created as part of our Advanced Topics Webinar, scheduled for March 26, 2021. This webinar focused on the high-throughput screening of surfaces using Exabyte.
# 
# In this notebook, we investigate the stability of a variety of different surfaces of Cu. After the surface energies are all automatically calculated, we finish by creating a Wulff Construction of Cu.

# In[]:


import os, sys

import ase.io
import ase.cluster

import pymatgen.ext.matproj
import pymatgen.io.ase
import pymatgen.symmetry.analyzer
import numpy as np

# Import settings file and utils
module_path = os.path.abspath(os.path.join('..'))
if module_path not in sys.path:
    sys.path.append(module_path)
from utils.generic import ensure_packages_are_installed, save_files
ensure_packages_are_installed()
from utils.material import get_all_slabs_and_terms
from utils.material import get_vasp_total_energy, get_slab_area, get_surface_energy
from settings import MATERIALS_PROJECT_API_KEY, ENDPOINT_ARGS, ORGANIZATION_ID

# Import relevant portions of the API client
from exabyte_api_client.endpoints.jobs import JobEndpoints
from exabyte_api_client.endpoints.projects import ProjectEndpoints
from exabyte_api_client.endpoints.materials import MaterialEndpoints
from exabyte_api_client.endpoints.workflows import WorkflowEndpoints
from exabyte_api_client.endpoints.bank_workflows import BankWorkflowEndpoints


# # Get Unit Cells
# 
# We begin by finding the following unit cell:
# - Cu, whose conventional unit cell is taken directly from Materials Project

# In[]:


def get_best_structure_by_formula(formula: str) -> pymatgen.ext.matproj.ComputedStructureEntry:
    # Create the Materials Project API object
    materials_project_api = pymatgen.ext.matproj.MPRester(MATERIALS_PROJECT_API_KEY)

    # Query for possible cells of Cu
    cu_structs = materials_project_api.get_entries(formula,
                                                   conventional_unit_cell=True,
                                                   sort_by_e_above_hull=True,
                                                   inc_structure="initial")
    best_cu_struct = cu_structs[0]
    return best_cu_struct

# Get Cu from Materials Project
cu_pymatgen = get_best_structure_by_formula("Cu").structure
cu_ase = pymatgen.io.ase.AseAtomsAdaptor.get_atoms(cu_pymatgen)


# Next, upload the unit cell to the Exabyte.IO Platform.
# 
# Note that this is in contrast to the way we import directly from Materials Project via IDs as in `materials/import_from_materials_project.ipynb` or in `material/api_interoperability_showcase.ipynb`. This is done deliberately to demonstrate the flexibility offered by the API.

# In[]:


# Create our Materials Endpoint
exabyte_materials_endpoint = MaterialEndpoints(*ENDPOINT_ARGS)

# Uplad Cu
cu_poscar_filename = "cu.vasp"
ase.io.write(cu_poscar_filename, cu_ase)
with open(cu_poscar_filename, "r") as inp:
    content = inp.read()
    cu_cell_material = exabyte_materials_endpoint.import_from_file(name="Copper_Unit_Cell", 
                                                                   content=content,
                                                                   owner_id=ORGANIZATION_ID)


# # Optimize the Unit Cell
# 
# We begin by specifying our INCAR file

# In[]:


# INCAR based on https://www.nature.com/articles/sdata201680#Sec2
incar_content = ["EDIFF = 1e-6",
                 "EDIFFG = -0.01",
                 "ISMEAR = 2",
                 "IBRION = 2",
                 "ISIF = 3",
                 "KPAR = 4",
                 "NSW = 300",
                 "ENCUT = 400"]

kpoints_content = ["Automatic mesh",
                   "0",
                   "Gamma",
                   "8 8 8",
                   "0 0 0"]


# Create a workflow based on the Variable-cell Relaxation workflow available [here](https://platform.exabyte.io/analytics/workflows/NAdKjws8qieKWeYnL)

# In[]:


# Start by finding a unit cell optimization workflow on the Exabyte platform. The bank workflow ID can be found in the bank workflows URL
bank_workflow_id = "NAdKjws8qieKWeYnL"

# Set up our Bank Workflow endpoint and copy over the workflow
exabyte_bank_workflows_endpoint = BankWorkflowEndpoints(*ENDPOINT_ARGS)
cu_workflow = exabyte_bank_workflows_endpoint.copy(bank_workflow_id,
                                                   account_id = ORGANIZATION_ID)

# Update the workflow with our custom INCAR file
vasp_unit = cu_workflow["subworkflows"][0]["units"][0]
for input_file in vasp_unit["input"]:
    if input_file["name"] == "INCAR":
        input_file['content'] = "\n".join(incar_content)
    elif input_file["name"] == "KPOINTS":
        input_file['content'] = "\n".join(kpoints_content)

# Set the names to something easy to recognize, and upload
cu_workflow['name'] = 'Copper_Cell_Relax'
exabyte_workflows_endpoint = WorkflowEndpoints(*ENDPOINT_ARGS)
cu_cellopt_workflow = exabyte_workflows_endpoint.create(cu_workflow)


# Create the job for the unit cell optimization

# In[]:


# Get the default projectID for the organization account (this can be substituted with a userID)
exabyte_projects_endpoint = ProjectEndpoints(*ENDPOINT_ARGS)
project_id = exabyte_projects_endpoint.list({"isDefault": True,
                                             "owner._id": ORGANIZATION_ID})[0]["_id"]

# Create the compute configuration for the jobs on Azure
exabyte_jobs_endpoint = JobEndpoints(*ENDPOINT_ARGS)
job_config = {"ppn": 4,
              "queue": "D",
              "nodes": 1,
              "time_limit": "00:10:00",
              "cluster": "cluster-001"}
compute = exabyte_jobs_endpoint.get_compute(**job_config)

# Create the Cu job
cu_job = exabyte_jobs_endpoint.create_by_ids([cu_cell_material],
                                             cu_cellopt_workflow["_id"],
                                             project_id,
                                             ORGANIZATION_ID,
                                             "cu_cellopt",
                                             compute)


# Finally, submit the job

# In[]:


exabyte_jobs_endpoint.submit(cu_job[0]["_id"])


# # Generate the Slabs

# We begin by extracting the relaxed unit cell from the job
# 

# In[]:


# Get a list of files for each
cu_job_id = cu_job[0]["_id"]

save_files(cu_job_id, exabyte_jobs_endpoint, "CONTCAR", "cu_relaxed.vasp")


# The below code uses Pymatgen to find all unique planes in a crystal. It then generates slabs with every possible termination in that plane. Finally, asymmetric slabs are filtered out, and the slabs are saved to a dictionary. The dictionary's format is: `{miller-index: {termination: {"slab": slab} }`. Note that here, to take advantage of several convenient functions in ASE, we have the function output ASE Atoms objects instead of PyMatGen objects.

# In[]:


cu_relaxed_pymatgen = pymatgen.core.structure.Structure.from_file("cu_relaxed.vasp")

# It is important to standardize the cell before cleaving planes
standardizer = pymatgen.symmetry.analyzer.SpacegroupAnalyzer(cu_relaxed_pymatgen)
cu_relaxed_pymatgen = standardizer.get_conventional_standard_structure()

cu_slabs = get_all_slabs_and_terms(cu_relaxed_pymatgen, thickness = 15, is_by_layers=False)


# # Write Slabs to POSCARs
# 
# A critical part of any VASP calculation is the actual POSCAR file - the file that contains the atomic coordinates.
# In this section, we take the slabs we generated above, and prepare them for the VASP calculation.

# In[]:


# Declare the vacuum size we want to have (for easy adjustment by future users)
vacuum_size = 15
for miller_index, term_dict in cu_slabs.items():
    for term, surface in term_dict.items():
        slab = surface["slab"]
        # Center the slab's coordinates and adjust the vacuum to vacuum_size
        # Note that the "vacuum" argument refers to the amount of vacuum on either side - so we divide by 2
        slab.center(vacuum=vacuum_size/2, axis=2)

        # Write the slab to a file
        formula = slab.get_chemical_formula(empirical=True)
        filename = f"{formula}_{miller_index}_term{term}.vasp"
        ase.io.write(filename, slab)


# # Upload POSCARS to Exabyte.io
# 
# Now that we have a set of prepared POSCARs, we can upload them to our user account on Exabyte. Let's create a couple material sets to hold these.

# In[]:


cu_materials_set = exabyte_materials_endpoint.create_set({"name" : "Webinar_Cu_Wulff_Slabs",
                                                         "owner": {"_id": ORGANIZATION_ID}})


# We want to keep track of material IDs as we go along, so we'll add this key to our Surface dictionary

# In[]:


for miller_index, term_dict in cu_slabs.items():
    for term, surface in term_dict.items():
        slab = surface["slab"]

        # Get the filename we generated above
        formula = slab.get_chemical_formula(empirical=True)
        material_name = f"{formula}_{miller_index}_term{term}"
        filename = f"{material_name}.vasp"

        # Import the material to Exabyte, and place it in the correct material set
        with open(filename, "r") as inp:
            content = inp.read()
        material_json = exabyte_materials_endpoint.import_from_file(name=material_name,
                                                                    content=content,
                                                                    owner_id=ORGANIZATION_ID)
        exabyte_materials_endpoint.move_to_set(material_json["_id"], "", cu_materials_set["_id"])

        # Adjust our dictionary to keep track of the Material ID
        surface["material_id"] = material_json["_id"]


# # Optimize the Cu Slabs
# 
# Now it is time to optimize the Cu slabs. We begin by specifying our INCAR file. It is largely the same as the previous INCAR, except with ISIF set to 2, to avoid changing the cell dimensions.

# In[]:


incar_content = ["EDIFF = 1e-6",
                 "EDIFFG = -0.01",
                 "ISMEAR = 2",
                 "IBRION = 2",
                 "ISIF = 2",
                 "KPAR = 4",
                 "NSW = 300",
                 "ENCUT = 400"]
kpoints_content = ["Automatic mesh",
                   "0",
                   "Gamma",
                   "8 8 1",
                   "0 0 0"]


# Next, we'll create the surface relaxation jobs. We start by cloning in a fixed cell relaxation workflow and adjusting it to suit our purposes.
# The workflow we will choose can be found [here](https://platform.exabyte.io/analytics/workflows/rTEtXntXo3ScGhi7q)

# In[]:


# Start by finding a fixed cell optimization workflow on the Exabyte platform.
bank_workflow_id = "rTEtXntXo3ScGhi7q"

# Copy over the workflow
workflow = exabyte_bank_workflows_endpoint.copy(bank_workflow_id,
                                                account_id=ORGANIZATION_ID)

# Add in our INCAR
vasp_unit = workflow["subworkflows"][0]["units"][0]
for input_file in vasp_unit["input"]:
    if input_file["name"] == "INCAR":
        input_file["content"] = "\n".join(incar_content)
    elif input_file["name"] == "KPOINTS":
        input_file["content"] = "\n".join(kpoints_content)

# Set the names to something easy to recognize, and upload
workflow['name'] = 'Copper_Slab_Relax'
workflow = exabyte_workflows_endpoint.create(workflow)


# Next, create the slab relaxation jobs

# In[]:


for miller_index, term_dict in cu_slabs.items():
    for term, surface in term_dict.items():
        # First, set up the compute parameters
        # Add an extra node if there are a lot of atoms in the cell
        n_nodes = int(np.rint(len(surface["slab"])/16))
        # Restrict the nodes to at least 1, but no more than 2
        n_nodes = min(max(n_nodes, 1), 2)

        job_config = {"ppn": 16,
                      "queue": "OF",
                      "nodes": n_nodes,
                      "time_limit": "12:00:00",
                      "cluster": "cluster-007"}
        compute = exabyte_jobs_endpoint.get_compute(**job_config)

        # Determine what we should name these jobs
        job_prefix = f"cu_slab_{miller_index}_Term{term}"

        # Get the material and workflow
        material_id = surface["material_id"]
        workflow_id = workflow["_id"]
        material = exabyte_materials_endpoint.get(material_id)

        # Set up the job and record the JOB ID, then submit the job
        cu_slab_job = exabyte_jobs_endpoint.create_by_ids([material],
                                                          workflow_id,
                                                          project_id,
                                                          ORGANIZATION_ID,
                                                          job_prefix,
                                                          compute)
        surface["job_id"] = cu_slab_job[0]["_id"]
        exabyte_jobs_endpoint.submit(surface["job_id"])


# # Extract Bulk and Slab Energy
# 
# 
# After our jobs complete, we can calculate the surface energy as follows:
# 
# (E_Slab - E_bulk * (N_Slab / N_Bulk)) / (2A)
# 
# We begin by extracting the total energy from the Cu unit cell

# In[]:


cu_job_id = cu_job[0]["_id"]
cu_bulk_energy = get_vasp_total_energy(cu_job_id, exabyte_jobs_endpoint)


# Then, we extract the total energy from each of the slab runs

# In[]:


for miller_index, term_dict in cu_slabs.items():
    for term, surface in term_dict.items():
        surface["slab_energy"] = get_vasp_total_energy(surface["job_id"], exabyte_jobs_endpoint)


# # Calculate the Surface Energy
# We'll iterate over each slab and calculate the surface energy

# In[]:


for miller_index, term_dict in cu_slabs.items():
    for term, surface in term_dict.items():
        slab = surface["slab"]
        # Slab and bulk energy
        e_slab = surface['slab_energy']
        e_bulk = cu_bulk_energy

        # N bulk and N Slab
        cu_cell = ase.io.read("cu_relaxed.vasp")
        n_bulk = len(cu_cell)
        n_slab = len(slab)

        # Slab Area
        vec_a = slab.cell[0]
        vec_b = slab.cell[1]
        area = get_slab_area(vec_a, vec_b)

        print(miller_index, e_slab, e_bulk, n_slab, n_bulk, area)

        # Surface Energy
        surface_energy = get_surface_energy(e_slab=e_slab, e_bulk=e_bulk,
                                            n_slab=n_slab, n_bulk=n_bulk,
                                            a=area)
        surface["surface_energy"] = surface_energy


# # Wulff Construction
# 
# Finally, now that we have a collection of surfaces and their energies, we can perform the Wulff Construction

# In[]:


surfaces = []
energies = []
for miller_index, term_dict in cu_slabs.items():
    # Turn the string miller index into a list, for ASE
    i,j,k = map(int, miller_index)
    plane = (i,j,k)
    surfaces.append(plane)
    # Find the best termination for the surface
    best_term_energy = np.inf
    for term, surface in term_dict.items():
        best_term_energy = min(best_term_energy, surface["surface_energy"])
    energies.append(best_term_energy)

cluster_size = 147
wulff = ase.cluster.wulff_construction("Cu", surfaces=surfaces,
                                       energies=energies, size=cluster_size,
                                       structure = "fcc")

ase.io.write("Cu_Wulff.xyz", wulff)


# In[]:


for surf, en in sorted(zip(surfaces, energies), key = lambda i: i[1]):
    print(surf, np.round(en,3))


# In[ ]:







from typing import Any, List, Optional, Union


def create_prompt_text(array: List[Any], element_name: str = "element", prompt_head: Optional[str] = None) -> str:
    """
    Create a prompt text for selecting an element from an array.
    Args:
        array (List[Any]): The array (list) of elements to select from.
        element_name (str): The name of the element to be used in the prompt text (e.g., "transformation", "interface")
        prompt_head: The prompt text to be displayed before the list of elements.

    Returns:
        str: The prompt text.
    """
    prompt_head = prompt_head or f"Select {element_name} by index:\n"
    prompt_body = "\n".join(f"{i}: {t}" for i, t in enumerate(array))
    return prompt_head + prompt_body


def get_integer_from_input(selected_index_str: str, array: List[Any]) -> Union[int, None]:
    """
    Get an integer from the input string and check if it is a valid index for the given array.
    Args:
        selected_index_str (str): The input string to convert to an integer.
        array (List[Any]): The array (list) of elements to select from.

    Returns:
        int: The selected index if it is valid, otherwise None.
    """
    try:
        selected_index = int(selected_index_str)
    except ValueError:
        print("Invalid input. Please enter a valid integer.")
        return None

    if selected_index < 0 or selected_index >= len(array):
        print("Invalid index.")
        return None
    return selected_index


def ui_prompt_select_array_element_by_index(
    array: List[Any], element_name: str = "element", prompt_head: Optional[str] = None
) -> Any:
    """
    Prompt the user to select an element from an array by index and return the chosen element.
    Args:
        array (List[Any]): The array (list) of elements to select from.
        element_name (str): The name of the element to be used in the prompt text (e.g., "transformation", "interface")
        prompt_head: The prompt text to be displayed before the list of elements.

    Returns:
        Any: The selected element from the array.
    """
    prompt_text = create_prompt_text(array, element_name, prompt_head)
    selected_index_str = input(prompt_text)
    index = get_integer_from_input(selected_index_str, array)
    if index is None:
        return None
    result = array[index]
    print(f"Selected {element_name}: ", array[index])
    return result


async def ui_prompt_select_array_element_by_index_pyodide(
    array: List[Any], element_name: str = "element", prompt_head: Optional[str] = None
) -> Any:
    """
    Prompt the user to select an element from an array by index and return the chosen element.
    Used in Pyodide environment, needs to be awaited since the return of input() is type of PyodideFuture.
    Args:
        array (List[Any]): The array (list) of elements to select from.
        element_name (str): The name of the element to be used in the prompt text (e.g., "transformation", "interface")
        prompt_head: The prompt text to be displayed before the list of elements.

    Returns:
        Any: The selected element from the array.
    """
    prompt_text = create_prompt_text(array, element_name, prompt_head)
    # `input()` in Pyodide returns PyodideFuture, which is not compatible with `int` in regular Python environment
    selected_index_str = await input(prompt_text)  # type: ignore
    index = get_integer_from_input(selected_index_str, array)
    if index is None:
        return None
    result = array[index]
    print(f"Selected {element_name}: ", array[index])
    return result



#!/usr/bin/env python
# coding: utf-8

# # Perform a High-Throughput Adsorption Study
# 
# This notebook was created to facilitate our April 2021 webinar, which detailed the creation of a simple model of CO adsorption. The video can be found in the following link:
# 
# https://www.youtube.com/watch?v=wOrqnR8UAbQ
# 
# The webinar's approach to placing CO molecules in a top-site configuration was based on the approach used in the following work:
# 
# Dean, J.; Taylor, M. G.; Mpourmpakis, G. Unfolding Adsorption on Metal Nanoparticles: Connecting Stability with catalysis. Sci Adv 2019, 5 (9), eaax5101. DOI: doi.org/10.1126/sciadv.aax5101
# 
# In this notebook example, we automate the placement of CO molecules to several Cu slabs. We then submit all of these adsorbed states to the cluster as DFT geometry optimization calculations using VASP.

# In[]:


import os, sys
import string

import ase.io
import ase.neighborlist
import ase.constraints

import numpy as np

# Import settings file and utils
module_path = os.path.abspath(os.path.join('..'))
if module_path not in sys.path:
    sys.path.append(module_path)
from utils import ensure_packages_are_installed
#ensure_packages_are_installed()
from material_utils import download_contcar
from settings import ENDPOINT_ARGS, ORGANIZATION_ID

# Import relevant portions of the API client
from exabyte_api_client.endpoints.jobs import JobEndpoints
from exabyte_api_client.endpoints.projects import ProjectEndpoints
from exabyte_api_client.endpoints.materials import MaterialEndpoints
from exabyte_api_client.endpoints.workflows import WorkflowEndpoints
from exabyte_api_client.endpoints.bank_workflows import BankWorkflowEndpoints


# # Query the Platform for the Slab
# 
# Get slab, acquire science

# In[]:


job_endpoint = JobEndpoints(*ENDPOINT_ARGS)
material_endpoint = MaterialEndpoints(*ENDPOINT_ARGS)

class JobData():
    """
    Dataclass for the job. Holds references to endpoints, the JSON for the job, along with
    other metadata such as the termination and miller index.
    
    Attributes:
        job_endpoint (JobEndpoints): A reference to the job endpoint from the API
        material_endpoint (MaterialEndpoints): A reference to the material endpoint from the API
        jobId (str): The ID of the job associated with an instance of this class
        job_json (dict): JSON data storing the job
        termination (str): Termination number for the slab.
        miller_index (str): Miller index that was used to cleave the slab.
        symbol (str): Atomic symbol for the element comprising the slab.
    """
    def __init__(self, jobId, job_endpoint, material_endpoint):
        self.job_endpoint = job_endpoint
        self.material_endpoint = material_endpoint
        
        self.jobId = jobId
        
        # Run the API call and get the job's name
        self.job_json = self.job_endpoint.get(jobId)
        self.name = self.job_json["name"]
        
        # Process the job name
        _, _, miller_index, end_string = self.name.split("_")
        termination_string, symbol = end_string.split(" ")
        self.termination = int("".join([char if char in string.digits else "" for char in termination_string]))
        self.miller_index = [int(index) for index in miller_index]
        self.symbol = symbol
        
    def __repr__(self):
        return(str(self.name))

# Here, we've stored all of our jobs in a specific set. We'll run a Mongo Query to find jobs
# that are inside of this set.
job_set_id = "WLBD7M9dZBaNH2mMX"
jobs = [JobData(job["_id"], job_endpoint, material_endpoint) for job in job_endpoint.list({"inSet":{"$elemMatch":{"_id":job_set_id}}})]


# # 

# # Download a Slab from the Platform
# 
# Now that we have our slab jobs, and a nice object to hold them, let's get them into ASE.

# ## Get an ASE oject

# In[]:


for job in jobs:
    job.filename = f"{'_'.join(map(str, job.miller_index))}_term{job.termination}.vasp"
    download_contcar(job.jobId, job.job_endpoint, job.filename)
    job.structure = ase.io.read(job.filename)


# ## Generate the ASE Object

# In[]:


# Ensure we've got at least a 5x5 angstrom surface, to help combat lateral interactions in the adsorbates
for job in jobs:
    a, b, c = job.structure.cell.lengths()
    while a < 5 or b < 5:
        if a < 5:
            job.structure *= [2,1,1]

        if b < 5:
            job.structure *= [1,2,1]
        
        a, b, c = job.structure.cell.lengths()


# ## Find the Surface Sites

# In[]:


import ase.neighborlist

def get_surface_indices(ase_structure, cns):
    surface_indices = []
    a, b, c = ase_structure.cell.lengths()
    for index, (atom, cn) in enumerate(zip(ase_structure, cns)):
        # Slabs are centered, so we can easily filter out one half of the atoms
        if cn < 12 and atom.position[2] > c/2:
            surface_indices.append(index)
    return surface_indices

for job in jobs:
    src, dest = ase.neighborlist.neighbor_list("ij", job.structure, 3)
    job.bonds = tuple(zip(src, dest))  
    job.cns = np.bincount(src)
    job.surface_indices = get_surface_indices(job.structure, job.cns)


# # CO Placement
# 
# Place CO on surface, surface norm is in Z-direction

# In[]:


def place_co_molecule(job, site):
    site_coords = job.structure[site].position
    carbon_atom = ase.Atom("C", site_coords + [0, 0, 1.43])
    oxygen_atom = ase.Atom("O", site_coords + [0, 0, 2*1.43])
    carbon_monoxide = ase.Atoms((carbon_atom, oxygen_atom))
    job.adsorption_states.append(job.structure + carbon_monoxide)
    
for job in jobs:
    job.adsorption_states = []
    for site in job.surface_indices:
        place_co_molecule(job, site)


# And finally, we'll freeze the surface to reduce the computational burden

# In[]:


def fix_all_slab_atoms(adsorption_state):
    constraint = ase.constraints.FixAtoms(indices = [atom.index for atom in adsorption_state if atom.symbol not in ("C", "O")])
    adsorption_state.set_constraint(constraint)

for job in jobs:
    for adsorption_state in job.adsorption_states:
        fix_all_slab_atoms(adsorption_state)


# # Adsorption Study
# 
# Conventionally, the adsorption energy is defined as an energetic comparison between the following two states:
# 1. The two species separated at an infinite distance.
# 2. The two species interacting with one another in some _complex_.
# 
# At an infinite distance (state 1), there is no interaction between the two species. As a result, the energy of the system is simply the sum of the energies of the two isolated systems. In our case (a molecule which has adsorbed to a slab), the isolated systems are the bare slab and the gas-phase molecule. 
# When the two species are combined (state 2), they interact with one another. The energetic quantification of this interaction (E_ads) is what we seek to determine.
# 
# Overall, the result is that we must perform three calculations: the gas-phase molecule (E_molecule), the bare slab (E_slab), and the adsorbed state (E_complex). The adsorption energy (E_ads) can then be calculated via the following equation.

# In[]:


get_ipython().run_cell_magic('latex', '', '\n\\begin{align}\nE_{ads}=E_{complex}-(E_{slab}+E_{molecule})\n\\end{align}')


# In[]:


for job in jobs:
    for index, adsorption_state in enumerate(job.adsorption_states):
        job.adsorption_state_names = []
        adsorption_state_name = job.filename.replace(".vasp", f"site_{index}")
        job.adsorption_state_names.append(adsorption_state_name) 
        ase.io.write(adsorption_state_name + ".vasp", adsorption_state)


# ## Create the Workflow

# In[]:


# Start by finding the slab optimization workflow from the previous webinar
bank_workflow_id = "ByGrAAkeGiopxJdyu"

# Copy over the workflow
bank_workflow_endpoint = BankWorkflowEndpoints(*ENDPOINT_ARGS)
workflow = bank_workflow_endpoint.copy(bank_workflow_id, account_id=ORGANIZATION_ID)

# Copy in our constraints
vasp_unit = workflow["subworkflows"][0]["units"][0]
for input_file in vasp_unit["input"]:
    if input_file["name"] == "POSCAR":
        input_file["content"] = "{{ input.POSCAR_WITH_CONSTRAINTS }}"

# Set the names to something easy to recognize, and upload
workflow_endpoint = WorkflowEndpoints(*ENDPOINT_ARGS)
workflow['name'] = 'Constrained Adsorbate_Relaxation'
workflow = workflow_endpoint.create(workflow)


# ## Upload the Materials

# In[]:


slab_materials_set = material_endpoint.create_set({"name" : "Webinar_Cu_CO_Slabs",
                                                            "owner": {"_id": ORGANIZATION_ID}})
for job in jobs:
    job.slab_ids = []
    for jobname in job.adsorption_state_names:
        with open(jobname + ".vasp", "r") as inp:
            content = "".join(inp.readlines())
        material_json = material_endpoint.import_from_file(name=jobname, content=content, owner_id=ORGANIZATION_ID)
        job.slab_ids.append(material_json["_id"])
        material_endpoint.move_to_set(material_json["_id"], "", slab_materials_set["_id"])


# ## Submit the Jobs

# In[ ]:


projects_endpoint = ProjectEndpoints(*ENDPOINT_ARGS)
project_id = projects_endpoint.list({"isDefault": True,
                                             "owner._id": ORGANIZATION_ID})[0]["_id"]      
        
for job in jobs:
    job.job_ids = []
    for jobname, adsorption_state, material_id in zip(job.adsorption_state_names,
                                                      job.adsorption_states,
                                                      job.slab_ids):
                # First, set up the compute parameters
        # Add an extra node if there are a lot of atoms in the cell
        n_nodes = int(np.rint(len(adsorption_state)))
        # Restrict the nodes to at least 1, but no more than 2
        n_nodes = min(max(n_nodes, 1), 2)

        job_config = {"ppn": 16,
                      "queue": "OF",
                      "nodes": n_nodes,
                      "time_limit": "12:00:00",
                      "cluster": "cluster-007"}
        compute = job.job_endpoint.get_compute(**job_config)
        
        workflow_id = workflow["_id"]
        material = material_endpoint.get(material_id)
        adsorbate_job = job.job_endpoint.create_by_ids([material],
                                                       workflow_id,
                                                       project_id,
                                                       ORGANIZATION_ID,
                                                       jobname,
                                                       compute)
        job.job_ids.append(adsorbate_job[0]["_id"])

    for job_id in job.job_ids:
        job.job_endpoint.submit(job_id)




import functools
import urllib.request
from typing import Any, Dict, Optional, Tuple, Union

import ase.build
import ase.constraints
import numpy as np
import pymatgen.core.surface
import pymatgen.io.ase
import pymatgen.symmetry.analyzer
from exabyte_api_client.endpoints.jobs import JobEndpoints


def download_file_by_name(job_id: str, job_endpoint: JobEndpoints, target: str, pattern: str) -> None:
    """
    Downloads a file from S3 and writes it to the local disk.

    Args:
        job_id (str): The ID string for the job that the file is to be downloaded from.
        job_endpoint (JobEndpoints): Job endpoint object from the Exabyte API Client
        target (str): Target filename that the file will be written to.
        pattern (str): The name of the file as it appears on the S3. For example, to download a file named
                       "CONTCAR", the pattern would be provided as "CONTCAR"
    Notes:
        This function does not check whether the file specified in `target` exists or not, and will always
        be overwritten it if the file is successfully downloaded.
    """
    # Get a list of files for each
    job_files = job_endpoint.list_files(job_id)

    # Find the file
    for file in job_files:
        if file["name"] == pattern:
            file_metadata = file

    # Get a download URL for the file
    file_signed_url = file_metadata["signedUrl"]

    # Download the file to memory
    server_response = urllib.request.urlopen(file_signed_url)

    # Write it to disk
    with open(target, "wb") as outp:
        outp.write(server_response.read())


download_contcar = functools.partial(download_file_by_name, pattern="CONTCAR")


# Pymatgen


def is_symmetric(slab: pymatgen.core.structure.Structure) -> bool:
    """
    Checks whether a slab is in a point group with inversion symmetry, which includes the following groups:
    -1, 2/m, mmm, 4/m, 4/mmm,-3, -3m, 6/m, 6/mmm, m-3, m-3m
    This technique is borrowed from the `nonstoichiometric_symmetrized_slab` method in PyMatGen's SlabGenerator.

    Args:
        slab (pymatgen.core.structure.Structure): Slab of interest

    Returns:
        True if the slab's spacegroup has inversion symmetry, otherwise False.
    """
    # Create a Spacegroup analyzer object with the slab
    spacegroup = pymatgen.symmetry.analyzer.SpacegroupAnalyzer(slab)
    # Check for inversion symmetry
    is_symmetric = spacegroup.is_laue()
    return is_symmetric


def get_all_slabs_and_terms(
    crystal: pymatgen.core.structure.Structure, thickness: Union[int, float], is_by_layers: bool
) -> Dict[str, Dict[str, Dict[str, Any]]]:
    """
    Gets all slabs and terminations for a given crystal, forcing a specific number of layers in the resultant slab.

    Args:
        crystal (pymatgen.core.structure.Structure):  Crystal of interest
        thickness (int or float): How thick the slab is supposed to be, by either angstroms or number of layers
        is_by_layers (bool): Whether thickness is by the number of layers or by angstroms

    Returns:
        Dict
    """
    # First, get a list of all non-symetrically-equivalent indices that the crystal has
    all_indices = pymatgen.core.surface.get_symmetrically_distinct_miller_indices(crystal, max_index=3)
    slabs = {}

    for plane in all_indices:
        slab_generator = pymatgen.core.surface.SlabGenerator(
            crystal, plane, min_slab_size=thickness, min_vacuum_size=10, center_slab=True, in_unit_planes=is_by_layers
        )

        # Generate all surface terminations, and add them to the list of slabs returned
        all_terminations = slab_generator.get_slabs()
        term_dict = {}
        symmetric_terminations = filter(is_symmetric, all_terminations)
        for term, surface in enumerate(symmetric_terminations):
            ase_surface = pymatgen.io.ase.AseAtomsAdaptor.get_atoms(surface)
            term_dict[str(term)] = {"slab": ase_surface}
        slabs["".join(map(str, plane))] = term_dict
    return slabs


# ASE


def get_bulk_bottom_and_top_frac_coords(slab: ase.Atoms, layers: int = 3) -> Tuple[float, float]:
    """
    Finds the top and bottom of the bulk, in fractional coordinates

    Args:
        slab (ase.Atoms): The slab of interest
        layers (int): How many layers are in the slab

    Returns:
        A tuple containing, in order, the fractional coordintes of the bottom and top of the bulk portion of the slab
    """
    # Determine how large the slab actually is
    c_direction_coords = [atom.scaled_position[2] for atom in slab]
    slab_range = max(c_direction_coords) - min(c_direction_coords)

    # Do some math to figure out where the bottom of the bulk starts
    layer_size = slab_range / layers
    bulk_bottom = min(c_direction_coords) + layer_size

    # Figure out where the top of the bulk starts
    bulk_top = min(c_direction_coords) + 2 * layer_size

    return (bulk_bottom, bulk_top)


def freeze_center_bulk(slab: ase.Atoms) -> None:
    """
    Applies an ASE FixAtoms constraint to atoms found in the center of the slab, in-place.

    Args:
        slab (ase.Atoms): The slab of interest

    Returns:
        None, this function changes the slab in-place.
    """
    # Get the fractional coordinates for the bottom and top of the bulk
    bulk_bottom, bulk_top = get_bulk_bottom_and_top_frac_coords(slab)

    # Filter to get the atoms between the bottom/top of the bulk - that is, the bulk atoms
    frozen_atoms = filter(lambda atom: bulk_bottom <= atom.scaled_position[2] <= bulk_top, slab)

    # Get the indices of the bulk atoms, and apply the constraint to the Atoms object
    frozen_atoms_indices = [atom.index for atom in frozen_atoms]
    fix_atoms_constraint = ase.constraints.FixAtoms(indices=frozen_atoms_indices)
    slab.set_constraint(fix_atoms_constraint)


def get_vasp_total_energy(job_id: str, jobs_endpoint: JobEndpoints) -> Optional[float]:
    """
    This function takes in a VASP Job ID, reads the OUTCAR, and returns the final energy reported in the run.

    Args:
        job_id (str): The ID of the job of interest
        jobs_endpoint (JobEndpoints): A job endpoint for interacting with the Exabyte platform

    Returns:
        The electronic energy reported by the VASP job or None
    """
    # Get the URL for the OUTCAR
    files = jobs_endpoint.list_files(job_id)
    file_metadata = None
    for file in files:
        if file["name"] == "OUTCAR":
            file_metadata = file

    # Get a download URL for each CONTCAR
    cell_outcar_signed_url = file_metadata["signedUrl"]  # type: ignore

    # Download the outcar to memory
    cell_response = urllib.request.urlopen(cell_outcar_signed_url)

    # And iterate through it, finding the last electronic energy reported
    outcar = cell_response.read().decode("utf-8")
    outcar = outcar.split("\n")
    unit_cell_energy = None
    for line in outcar:
        if "sigma->0" in line:
            unit_cell_energy = float(line.strip().split()[-1])
    return unit_cell_energy


def get_surface_energy(e_slab: float, e_bulk: float, n_slab: float, n_bulk: float, a: float) -> float:
    """
    Calculates the slab energy according to the following formula:
    (E_Slab - E_bulk * (N_Slab / N_Bulk)) / (2A)
    """
    surface_energy = (e_slab - e_bulk * (n_slab / n_bulk)) / (2 * a)
    return surface_energy


def get_slab_area(a_vector: np.ndarray, b_vector: np.ndarray) -> float:
    """
    Gets the area of a slab defined by the two unit vectors.
    The magnitude of the cross product of the vectors is the area of the parallelogram they enclose.

    Args:
        a_vector
    """
    crossprod = np.cross(a_vector, b_vector)
    magnitude = np.linalg.norm(crossprod)
    return magnitude



import io
from typing import Dict, List, Optional, Union

import ipywidgets as widgets
from ase.build import make_supercell
from ase.io import write
from IPython.display import display
from mat3ra.made.material import Material
from mat3ra.made.tools.convert import to_ase
from mat3ra.utils.array import convert_to_array_if_not
from pydantic import BaseModel


def get_material_image(material: Material, title: str, rotation="0x,0y,0z", repetitions=[1, 1, 1]):
    """
    Returns an image of the material structure with the specified title.

    Args:
        material (Material): Material object to visualize.
        title (str): Title of the image.
        rotation (str): Rotation of the image, in degrees around the x, y, and z axes (e.g., "-90x,90y,0z").
        repetitions (list): Repetitions alongside a,b,c lattice vectors.

    Returns:
        tuple: Tuple containing the image and the title.
    """

    ase_atoms = to_ase(material)
    # Create supercell for visualization
    supercell_matrix = [[repetitions[0], 0, 0], [0, repetitions[1], 0], [0, 0, repetitions[2]]]
    material_repeat = make_supercell(ase_atoms, supercell_matrix)
    text = f"{ase_atoms.get_chemical_formula()} - {title} - rotation: {rotation}"

    # Write image to a buffer to display in HTML
    buf = io.BytesIO()
    write(buf, material_repeat, format="png", rotation=rotation)
    buf.seek(0)
    return buf.read(), text


def create_image_widget(image_data, format="png", object_fit="contain"):
    """
    Creates an Image widget with specified layout settings.

    Args:
        image_data (bytes): The image data to be displayed.
        format (str): The format of the image, default is 'png'.
        object_fit (str): CSS object-fit property value, default is 'contain'.

    Returns:
        widgets.Image: A configured image widget.
    """
    image = widgets.Image(value=image_data, format=format)
    image.layout.object_fit = object_fit
    return image


def create_responsive_image_grid(image_tuples, max_columns=3):
    """
    Create a responsive image grid that can display images from a specified folder.
    Ensures images are displayed at their true sizes and limits the grid to a maximum of three columns.

    Args:
        image_tuples (list): List of tuples where each tuple contains an image and a title.
        max_columns (int): Maximum number of columns in the grid.
    """
    items = [
        widgets.VBox(
            [
                widgets.Label(value=title, layout=widgets.Layout(height="30px", align_self="center")),
                create_image_widget(image, object_fit="contain"),
            ],
            layout=widgets.Layout(align_items="center", padding="0px 0px 10px 0px"),
        )  # Adjust padding as needed
        for image, title in image_tuples
    ]

    column_width = f"minmax(100px, {100 / max_columns}%)"

    grid = widgets.GridBox(
        items,
        layout=widgets.Layout(
            grid_template_columns=f"repeat({max_columns}, {column_width})",
            grid_gap="10px",
            width="100%",
        ),
    )
    return grid


class MaterialViewProperties(BaseModel):
    repetitions: List[int] = [1, 1, 1]
    rotation: str = "0x,0y,0z"
    title: str = "Material"


def visualize_materials(
    materials_to_view: Union[List[Material], List[Dict[str, Union[Material, dict]]]],
    repetitions: Optional[List[int]] = [1, 1, 1],
    rotation: Optional[str] = "0x,0y,0z",
    title: Optional[str] = "Material",
) -> None:
    """
    Visualize the material(s) in the output cell.
    Args:
        materials_to_view: Mist of Materials or a list of dictionaries:
        {"material": Material, "title": str,"repetitions": List[int], "rotation": str}.
        repetitions (Optional[List[int]]): Repetitions alongside a, b, c lattice vectors.
        rotation (Optional[str]): Rotation of the image, in degrees around the x, y, and z axes (e.g., "-90x,90y,0z").
        title (Optional[str]): Title of the image.

    Returns:
        None
    """
    materials_to_view = convert_to_array_if_not(materials_to_view)

    if not materials_to_view:
        print("No materials to visualize.")
        return

    default_properties = MaterialViewProperties(
        title=title if title is not None else MaterialViewProperties.title,
        repetitions=repetitions if repetitions is not None else MaterialViewProperties.repetitions,
        rotation=rotation if rotation is not None else MaterialViewProperties.rotation,
    )

    items = []
    for material_entry in materials_to_view:
        if isinstance(material_entry, Material):
            material = material_entry
            properties = default_properties
        elif isinstance(material_entry, dict) and "material" in material_entry:
            material = material_entry["material"]
            properties = MaterialViewProperties(
                title=material_entry.get("title", default_properties.title),
                repetitions=material_entry.get("repetitions", default_properties.repetitions),
                rotation=material_entry.get("rotation", default_properties.rotation),
            )
        else:
            print("Invalid material entry:", material_entry)
            continue

        image_data, image_title = get_material_image(
            material, title=properties.title, rotation=properties.rotation, repetitions=properties.repetitions
        )
        items.append((image_data, image_title))

    display(create_responsive_image_grid(items))
